num_of_snippets_to_query = 10
max_num_of_snippets = 5

INSTRUCTIONS_FOR_REVIEW = """\
> [!TIP]
> To get Sweep to edit this pull request, you can:
> * Comment below, and Sweep can edit the entire PR
> * Comment on a file, Sweep will only modify the commented file
> * Edit the original issue to get Sweep to recreate the PR from scratch"""
